//
// Board.h - Encapsulation of a chess board
//
// Purpose: This object contains all of the data and methods required to
// process a chess board in the game.  It uses the ubiquitous "bitboard"
// representation.
//
// History:
// 08.06.00 Created ( F. D. Larame )
// 14.08.00 Made "HashLock" a relative clone of "HashKey"; the java
//          Object.hashCode method is unsuitable to our purposes after all,
//          probably because it includes memory addresses in the calculation.
//          ( F. D. Larame )
// 19.03.03 Converted to C++ ( C. Kutz )
//
//////////////////////////////////////////////////////////////////////

#if !defined( AFX_BOARD_H__AC9A029C_8849_40E6_9BE2_BE23065EEDE8__INCLUDED_ )
#define AFX_BOARD_H__AC9A029C_8849_40E6_9BE2_BE23065EEDE8__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Move.h"

/**
 * public class Board
 *
 * Notes:
 * 1. The squares are numbered line by line, starting in the corner occupied by
 * Black's Queen's Rook at the beginning of the game.  There are no constants
 * to represent squares, as they are usually manipulated algorithmically, in
 * sequences, instead of being explicitly identified in the code.
 */
class Board  
{


public:
/** Codes representing pieces */
	static const int PAWN;
	static const int KNIGHT;
	static const int BISHOP;
	static const int ROOK;
	static const int QUEEN;
	static const int KING;
	static const int WHITE_PAWN;
	static const int WHITE_KNIGHT;
	static const int WHITE_BISHOP;
	static const int WHITE_ROOK;
	static const int WHITE_QUEEN;
	static const int WHITE_KING;
	static const int BLACK_PAWN;
	static const int BLACK_KNIGHT;
	static const int BLACK_BISHOP;
	static const int BLACK_ROOK;
	static const int BLACK_QUEEN;
	static const int BLACK_KING;
	static const int EMPTY_SQUARE;

  // Useful loop boundary constants, to allow looping on all bitboards and
  // on all squares of a chessboard
	static const int ALL_PIECES;
	static const int ALL_SQUARES;

  // Indices of the "shortcut" bitboards containing information on "all black
  // pieces" and "all white pieces"
	static const int ALL_WHITE_PIECES;
	static const int ALL_BLACK_PIECES;
	static const int ALL_BITBOARDS;

  // The possible types of castling moves; add the "side" constant to
  // pick a specific move for a specific player
	static const int CASTLE_KINGSIDE;
	static const int CASTLE_QUEENSIDE;

private:
  // Private table of random numbers used to compute Zobrist hash values.
  // Contains a signature for any kind of piece on any square of the board
  static const int HashKeyComponents[ 12 * 64 ];  // !! two-dimensional array !!
  static const int HashLockComponents[ 12 * 64 ]; // !! two-dimensional array !!


public:
	// An array of bitfields, each of which contains the single bit associated
  // with a square in a bitboard
  static __int64 *SquareBits;

  static __int64 EXTRAKINGS_WHITE_KINGSIDE;
  static __int64 EXTRAKINGS_WHITE_QUEENSIDE;
  static __int64 EXTRAKINGS_BLACK_KINGSIDE;
  static __int64 EXTRAKINGS_BLACK_QUEENSIDE;
  static __int64 EMPTYSQUARES_WHITE_KINGSIDE;
  static __int64 EMPTYSQUARES_WHITE_QUEENSIDE;
  static __int64 EMPTYSQUARES_BLACK_KINGSIDE;
  static __int64 EMPTYSQUARES_BLACK_QUEENSIDE;

	static int *PieceValues;

  // Private table of tokens ( string representations ) for all pieces 
  static string *PieceStrings;
		
  // And a few flags for special conditions.  The ExtraKings are a device
  // used to detect illegal castling moves: the rules of chess forbid castling
  // when the king is in check or when the square it flies over is under
  // attack; therefore, we add "phantom kings" to the board for one move only,
  // and if the opponent can capture one of them with its next move, then
  // castling was illegal and search can be cancelled
  __int64 ExtraKings[ 2 ];
  
  // Data needed to compute the evaluation function
  int MaterialValue[ 2 ];
  
	int NumPawns[ 2 ];

	/** Whose turn is it? */
	int CurrentPlayer;

	int MoveNo;

private:
	/**
	  * The actual data representation of a chess board.  First, an array of
    * bitboards, each of which contains flags for the squares where you can
    * find a specific type of piece
    */
	__int64 BitBoards[ 14 ];  // 14 == ALL_BITBOARDS

  /** Stores CastlingStatus
		* 
		*/
  bool	CastlingStatus[ 4 ];
  
	bool HasCastled[ 2 ];
  
	__int64 EnPassantPawn;
		
public:
	void ShowInfo();
	static void InitStatics();
	static void DeleteStatics();
	
	static int  CoordToSquare( const string& coord );
	static int  CoordToSquare( const char* const coord );
	static bool CoordToSquare( const string& coords , string* squares );
	static bool CoordToSquare( const string& coords, int* sourceSquare, int* destSquare );
	static bool CoordToSquare( const string& coords, int* sourceSquare, int* destSquare, char* pawnPromote ); 
	static bool SquareToCoord( const int whichSquare, string* coord );
	static bool SquareToCoord( const int whichSourceSquare, const int whichDestSquare, string* coord );
	static bool SquareToCoord( const string& whichSquares, string* coord );

	Board();
	virtual ~Board();

	__int64 GetBitBoard( int which ) const { return BitBoards[ which ]; }
	bool GetCastlingStatus( int which ) const { return CastlingStatus[ which ]; }
	int  GetCurrentPlayer()  const { return CurrentPlayer; }
	__int64 GetEnPassantPawn() const { return EnPassantPawn; }
	__int64 GetExtraKings( int side ) const { return ExtraKings[ side ]; }
    bool GetHasCastled( int which ) const { return HasCastled[ which ]; }
	
	void SetExtraKings( int side, __int64 val );
	void ClearExtraKings( int side );
  
  // Look for the piece located on a specific square
	int FindBlackPiece( int square );
	int FindWhitePiece( int square );
	int FindPiece( int square, int* Side );

	/**
      * Make a deep copy of a Board object; assumes that memory has already
      * been allocated for the new object, which is always true since we
      * "allocate" Boards from a permanent array
      */
	bool Clone( Board *source );
	/**
      * Display the board on standard output
      */
	bool Print();
	/** 
	  * Change the identity of the player to move
	  */
	int SwitchSides();
	/** 
	  * Compute a 32-bit integer to represent the board, according to Zobrist[ 70 ] 
	  */
	__int64 HashKey();
	/**
	  * Compute a second 32-bit hash key, using an entirely different set
	  * piece/square components.
	  * This is required to be able to detect hashing collisions without
	  * storing an entire Board in each slot of the TranspositionTable,
	  * which would gobble up inordinate amounts of memory
	  */
	int HashLock();
	int HashLock2();
	/**
    * Change the Board's internal representation to reflect the move
    * received as a parameter
    */
	bool ApplyMove( Move *theMove );
	
	/**
    * Compute the board's material balance, from the point of view of the "side"
    * player.  This is an exact clone of the eval function in CHESS 4.5
    */
	int EvalMaterial( int side );
	/** 
	  * Restore the board to a game-start position 
	  */
	bool StartingBoard();
	/**
	  * Load the board position from a file
		*/
	bool Load( char *filename, Board* newboard );

	istream& read( istream &is ) const;
	
	ostream& write( ostream &os ) const;

private:
	/** 
	  * Place a specific piece on a specific board square 
	  */
	bool AddPiece( int whichSquare, int whichPiece );
	/**
	 * Eliminate a specific piece from a specific square on the board
	 * Note that you MUST know that the piece is there before calling this,
	 * or the results will not be what you expect!
	 */
	bool RemovePiece( int whichSquare, int whichPiece );
	/** 
	  * Remove every piece from the board 
	  */
	bool EmptyBoard();
	/**
	  * Change one of the "castling status" flags
	  * parameter whichFlag should be a sum of a side marker and a castling
	  * move identifier, for example, Player.SIDE_WHITE + CASTLE_QUEENSIDE
	  */
	bool SetCastlingStatus( int whichFlag, bool newValue );
	/**
	  * If a pawn move has just made en passant capture possible, mark it as
	  * such in a bitboard ( containing the en passant square only )
	  */
	bool SetEnPassantPawn( int square );
	bool SetEnPassantPawn( __int64 bitboard );
	/**
	  * Indicates that there is no en passant square at all.  Technically, this
	  * job could have been handled by SetEnPassaantPawn( __int64 ) with a null
	  * parameter, but I have chosen to add a method to avoid problems if I ever
	  * forgot to specify 0L: using 0 would call the first form of the Set method
	  * and indicate an en passant pawn in a corner of the board, with possibly
	  * disastrous consequences!
	  */
	bool ClearEnPassantPawn();
    /** 
	  * Whose turn is it? 
	  */
	bool SetCurrentPlayer( int which );
};

#endif // !defined( AFX_BOARD_H__AC9A029C_8849_40E6_9BE2_BE23065EEDE8__INCLUDED_ )
