//
// CutoffHistory - A heuristic used to pick an order of evaluation for moves
//
// The history heuristic is an extension of the old "killer move" system: if
// a move has caused a lot of cutoffs recently, it will be tried early in the
// hope that it will do so again.
//
// Using the history table is a gamble.  We could do without it entirely,
// compute "successor positions" for each possible moves, look them up in
// the transposition table and hope to get a cutoff this way, which would
// insure fast cutoffs whenever possible.  On the other hand, HistoryTable
// has no knowledge of the contents of the transposition table, so it may
// cause a deep search of several moves even though another one would result
// in an immediate cutoff...  However, History requires far less memory
// and computation than creating a ton of successor Board objects, so we
// hope that, on average, it will still be more efficient overall.
//
/////////////////////////////////////////////////////////////////////////////////

#if !defined( AFX_CUTOFFHISTORY_H__38235B45_2993_423A_A939_2C29A20FF9C1__INCLUDED_ )
#define AFX_CUTOFFHISTORY_H__38235B45_2993_423A_A939_2C29A20FF9C1__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


class CutoffHistory  
{


/***************************************************************************
 * Static Data Members
 **************************************************************************/
private:
	/** This is a singleton class; the same history can be shared by two AI's */
	static CutoffHistory *theInstance;


/***************************************************************************
 * Data Members
 **************************************************************************/
public:
	// the table itself; a separate set of cutoff counters exists for each
	// side
	int History[ 2 ][ 64 ][ 64 ];  // History[ Player ][ source ][ dest ]
	static int *pCurrentHistory; // pCurrenthistory[ source ][ dest ]


/***************************************************************************
 * Constructor / Destructor
 **************************************************************************/
	CutoffHistory();


/***************************************************************************
 * Static Accessors
 **************************************************************************/
public:
	static CutoffHistory *GetInstance();

/***************************************************************************
 * Operations
 **************************************************************************/
public:
	void Print();
	/**
	  * History table compilation 
	  */
	bool AddCount( int whichPlayer, Move *mov, int depth = 1 );
	/**
	 * Once in a while, we must erase the history table to avoid ordering
	 * moves according to the results of very old searches
	 */
	bool Forget();
	/** 
	 * Sort a list of moves, the key is the Historytable! 
	 * This sort-algorithm has to be stable!
	 */
	bool SortMoveList( MoveListGenerator *theList, int movingPlayer );
	istream& read( istream &is ) const;
	ostream& write( ostream &os ) const;
};

#endif // !defined( AFX_CUTOFFHISTORY_H__38235B45_2993_423A_A939_2C29A20FF9C1__INCLUDED_ )
