//
// WinBoardConnection - A Class to communicate with WinBoard
//
// WinBoard is a gui for chess programs. This class is built for version 2 
// of WinBoard protocol. WinBoard uses standard input and output, so normal
// console output is defected to a file.
//
//////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include "WinBoardConnection.h"

/***************************************************************************
 * Static Class members
 **************************************************************************/

 const int WinBoardConnection::CMD_NONE			=  0;
 const int WinBoardConnection::CMD_UNKNOWN		=  1;
 const int WinBoardConnection::CMD_NEW			=  2;
 const int WinBoardConnection::CMD_USERMOVE		=  3;
 const int WinBoardConnection::CMD_PROTOVER		=  4;
 const int WinBoardConnection::CMD_XBOARD		=  5;
 const int WinBoardConnection::CMD_QUIT			=  6;
 const int WinBoardConnection::CMD_FEATURE		=  7;
 const int WinBoardConnection::CMD_ILLEGALMOVE	=  8;
 const int WinBoardConnection::CMD_ERROR		=  9;
 const int WinBoardConnection::CMD_MOVE			= 10;
 const int WinBoardConnection::CMD_RESIGN		= 11;
 const int WinBoardConnection::CMD_THINKING		= 12;
 const int WinBoardConnection::CMD_REMOVE		= 13;
 const int WinBoardConnection::CMD_WHITE		= 14;
 const int WinBoardConnection::MAX_CMD			= 20;
 bool WinBoardConnection::changeAI = false;
 bool WinBoardConnection::firstMove = true;
 
 WinBoardConnection* WinBoardConnection::self = NULL;
 const string WinBoardConnection::traceConnFilename = "ConnTrace.txt";
 bool WinBoardConnection::isActive = false;
 
/***************************************************************************
 * Constructor / Destructor
 **************************************************************************/
WinBoardConnection::WinBoardConnection()
{
	isActive = true;
	self = this;

	commandStrings = new string[ MAX_CMD ];


	commandStrings[ CMD_NONE ]			= "";
	commandStrings[ CMD_UNKNOWN ]		= "";
	commandStrings[ CMD_NEW ]			= "new";
	commandStrings[ CMD_USERMOVE ]		= "usermove";
	commandStrings[ CMD_PROTOVER ]		= "protover";
	commandStrings[ CMD_XBOARD ]		= "xboard";
	commandStrings[ CMD_QUIT ]			= "quit";
	commandStrings[ CMD_THINKING ]		= "thinking";
	commandStrings[ CMD_WHITE ]			= "white";

	commandStrings[ CMD_FEATURE ]		= "feature";
	commandStrings[ CMD_ILLEGALMOVE ]	= "Illegal move";
	commandStrings[ CMD_ERROR ]			= "Error";
	commandStrings[ CMD_MOVE ]			= "move";
	commandStrings[ CMD_RESIGN ]		= "resign";
	commandStrings[ CMD_REMOVE ]		= "remove";

	traceConn.open( traceConnFilename.c_str(), ios::out | ios::unitbuf );
	traceConn << "Trace started...\n\n";

	string par;
	int cmdtype;
	ScanInput( &cmdtype, &par );
	if ( cmdtype != CMD_XBOARD ) {
		cerr << "'xboard' command expected";
		exit( 0 );
	}
	ScanInput( &cmdtype, &par );
	if ( cmdtype != CMD_PROTOVER ) {
		cerr << "'protover' command expected";
		exit( 0 );
	}
	
	// tell X/Winboard the features of our implementation
	SendFeatures();
}

WinBoardConnection::~WinBoardConnection()
{
	isActive = false;

	traceConn.close();
	
	delete [] commandStrings;
}

/***************************************************************************
 * Operations
 **************************************************************************/

int WinBoardConnection::getPlayerType( int Side )
{
	if ( Side == Player::SIDE_WHITE ) return Game::HUMANWINBOARD; 
	return Game::AIWINBOARD;
}

string WinBoardConnection::getPlayerName( int Side )
{
	if ( Side == Player::SIDE_WHITE ) return "White - Human - WinBoard"; 
	return "Black - AI - WinBoard";
}

/* ScanInput parses the command string 'input' and returns a command type
   'cmdtype' with its parameters 'par */
bool WinBoardConnection::ScanInput( int *whichCommand, string *par )
{
	string input;

	// get characters from stdin
	getCommand( &input );
	
	*whichCommand = CMD_UNKNOWN;
	*par = input;
	
	if ( ( input == "" ) || ( input == "\n" ) ) return false;

	// scan input for all known commands
	for ( int i = 0; i < MAX_CMD; i++ ) {
		if ( commandStrings[ i ] == "" ) continue;
		if ( commandStrings[ i ] == input.substr( 0, commandStrings[ i ].size() ) ) {
			*whichCommand = i;
			if ( commandStrings[ i ].size() + 1 >= input.size() ) {
				*par = ""; 
			} else *par = input.substr( commandStrings[ i ].size() + 1, input.size() - 2 );
			break;
		}
	}

	if ( ( *whichCommand == CMD_NONE ) || ( *whichCommand == CMD_UNKNOWN ) ) return false;

	return true;
}


/* getCommand waits for input of standart input */
void WinBoardConnection::getCommand( string* cmd ) {
	string line;		
	char ch;
	line = "";
	do {
		cin.get( ch );
		switch ( ch ) {
			case '\n':	break;
						
			default:	line += tolower( ch );
			
		}
	} while ( ch != '\n' );

	*cmd = line;

	traceConn << "In: " << line << "\n";
	traceConn.flush();
}

/* send is used for standart output - to send commands to Winboard */
void WinBoardConnection::send( int whichCommand, string par )
{
	traceConn << "Out: " << commandStrings[ whichCommand ] << " " << par << "\n";
	traceConn.flush();

	if ( par != "" ) par = " " + par;
	*Output::cout << commandStrings[ whichCommand ] << par << endl;
	if ( whichCommand != CMD_THINKING ) {		// << thinking has no commandstring - just parameters
		cout << commandStrings[ whichCommand ] << par << endl;
	} else cout << par.substr( 1 ) << endl;

	cout.flush();
}

/* SendFeature command is used to set the features this program supports. This is requiered for Winboard */
void WinBoardConnection::SendFeatures() {
	cout << "feature";
	traceConn << "Out: feature";

	/*
	ping ( boolean, default 0, recommended 1 ) 
	If ping=1, xboard may use the protocol's new "ping" command; if ping=0, xboard will not use the command. 
	*/
	const int ping = 0; 
	cout << " ping=" << ping;
	traceConn << " ping=" << ping;

	/*
	setboard ( boolean, default 0, recommended 1 ) 
	If setboard=1, xboard will use the protocol's new "setboard" command to set up positions; if setboard=0, it will use the older "edit" command. 
	playother ( boolean, default 0, recommended 1 ) 
	*/
	const int setboard = 1;
	cout << " setboard=" << setboard;
	traceConn << " setboard=" << setboard;

	/*
	If playother=1, xboard will use the protocol's new "playother" command when appropriate; if playother=0, it will not use the command. 
	*/
	const int playother = 0;
	cout << " playother=" << playother;
	traceConn << " playother=" << playother;

	/*
	san ( boolean, default 0 ) 
	If san=1, xboard will send moves to the engine in standard algebraic notation ( SAN ); for example, Nf3. If san=0, xboard will send moves in coordinate notation; for example, g1f3. See MOVE in section 8 above for more details of both kinds of notation. 
	*/
	const int san = 0;
	cout << " san=" << san;
	traceConn << " san=" << san;

	/*
	usermove ( boolean, default 0 ) 
	If usermove=1, xboard will send moves to the engine with the command "usermove MOVE"; if usermove=0, xboard will send just the move, with no command name. 
	*/
	const int usermove = 1;
	cout << " usermove=" << usermove;
	traceConn << " usermove=" << usermove;

	/*
	time ( boolean, default 1, recommended 1 ) 
	If time=1, xboard will send the "time" and "otim" commands to update the engine's clocks; if time=0, it will not. 
	*/
	const int time = 0;
	cout << " time=" << time;
	traceConn << " time=" << time;

	/*
	draw ( boolean, default 1, recommended 1 ) 
	If draw=1, xboard will send the "draw" command if the engine's opponent offers a draw; if draw=0, xboard will not inform the engine about draw offers. Note that if draw=1, you may receive a draw offer while you are on move; if this will cause you to move immediately, you should set draw=0. 
	*/
	const int draw = 0;
	cout << " draw=" << draw;
	traceConn << " draw=" << draw;

	/*
	sigint ( boolean, default 1 ) 
	If sigint=1, xboard may send SIGINT ( the interrupt signal ) to the engine as section 7 above; if sigint=0, it will not. 
	*/
	const int sigint = 0;
	cout << " sigint=" << sigint;
	traceConn << " sigint=" << sigint;

	/*
	sigterm ( boolean, default 1 ) 
	If sigterm=1, xboard may send SIGTERM ( the termination signal ) to the engine as section 7 above; if sigterm=0, it will not. 
	*/
	const int sigterm = 0;
	cout << " sigterm=" << sigterm;
	traceConn << " sigterm=" << sigterm;

	/*
	reuse ( boolean, default 1, recommended 1 ) 
	If reuse=1, xboard may reuse your engine for multiple games. If reuse=0 ( or if the user has set the -xreuse option on xboard's command line ), xboard will kill the engine process after every game and start a fresh process for the next game. 
	*/
	const int reuse = 0;
	cout << " reuse=" << reuse;
	traceConn << " reuse=" << reuse;

	/*
	analyze ( boolean, default 1, recommended 1 ) 
	If analyze=0, xboard will not try to use the "analyze" command; it will pop up an error message if the user asks for analysis mode. If analyze=1, xboard will try to use the command if the user asks for analysis mode. 
	*/
	const int analyze = 1;
	cout << " analyze=" << analyze;
	traceConn << " analyze=" << analyze;

	/*
	myname ( string, default determined from engine filename ) 
	This feature lets you set the name that xboard will use for your engine in window banners, in the PGN tags of saved game files, and when sending the "name" command to another engine. 
	*/
	const string myname = "\"BaChess " + Game::Version + "\"";
	cout << " myname=" << myname;
	traceConn << " myname=" << myname;

	/*
	variants ( string, see text below ) 
	This feature indicates which chess variants your engine accepts. It should be a comma-separated list of variant names. See the table under the "variant" command in section 8 above. If you do not set this feature, xboard will assume by default that your engine supports all variants. ( However, the -zippyVariants command-line option still limits which variants will be accepted in Zippy mode. ) It is recommended that you set this feature to the correct value for your engine ( just "normal" in most cases ) rather than leaving the default in place, so that the user will get an appropriate error message if he tries to play a variant that your engine does not support. 
	*/
	const string variants = "\"losers\"";
	cout << " variants=" << variants;
	traceConn << " variants=" << variants;

	/*
	colors ( boolean, default 1, recommended 0 ) 
	If colors=1, xboard uses the obsolete "white" and "black" commands in a stylized way that works with most older chess engines that require the commands. See the "Idioms" section below for details. If colors=0, xboard does not use the "white" and "black" commands at all. 
	*/
	const int color = 0;
	cout << " color=" << color;
	traceConn << " color=" << color;

	/*
	ics ( boolean, default 0 ) 
	If ics=1, xboard will use the protocol's new "ics" command to inform the engine of whether or not it is playing on a chess server; if ics=0, it will not. 
	*/
	const int ics = 0;
	cout << " ics=" << ics;
	traceConn << " ics=" << ics;

	/*
	name ( boolean, see text below ) 
	If name=1, xboard will use the protocol's "name" command to inform the engine of the opponent's name; if name=0, it will not. By default, name=1 if the engine is playing on a chess server; name=0 if not. 
	*/
	const int name = 0;
	cout << " name=" << name;
	traceConn << " name=" << name;

	/*
	pause ( boolean, default 0 ) 
	If pause=1, xboard may use the protocol's new "pause" command; if pause=0, xboard assumes that the engine does not support this command. 
	*/
	const int pause = 0;
	cout << " pause=" << pause;
	traceConn << " pause=" << pause;

	/*
	done ( integer, no default ) 
	If you set done=1 during the initial two-second timeout after xboard sends you the "xboard" command, the timeout will end and xboard will not look for any more feature commands before starting normal operation. If you set done=0, the initial timeout is increased to one hour; in this case, you must set done=1 before xboard will enter normal operation. 
	*/
	const int done = 1;
	cout << " done=" << done;
	traceConn << " done=" << done;

	cout << "\n";
	traceConn << "\n";
}

WinBoardConnection* WinBoardConnection::getInstance()
{
	return self;
}

